/*
 * File: Snowfall.java
 * ================================================================
 * A program that simulates falling snow on a scene.
 */
import acm.program.*;
import acm.graphics.*;
import acm.util.*;
import java.awt.*;
import java.awt.event.*;

public class Snowfall extends GraphicsProgram {
	/* Preferred window size. */
	public static final int APPLICATION_WIDTH = 500;
	
	/* The delay between frames. */
	private static final double PAUSE_TIME = 1000.0 / 100;
	
	/* How fast the snowflakes fall. */
	private static final double Y_VELOCITY = 5;
	
	/* The color of a snowflake. */
	private static final Color SNOWFLAKE_COLOR = Color.GRAY;
	
	/* The size of a snowflake. */
	private static final double SNOWFLAKE_SIZE = 20;
	
	/* The size of a block. */
	private static final double BLOCK_WIDTH = 100;
	private static final double BLOCK_HEIGHT = 25;
	
	public void run() {
		addMouseListeners();
		while (true) {
			dropOneSnowflake();
		}
	}
	
	/**
	 * Chooses a random X coordinate for a snowflake.
	 * 
	 * @return A random X coordinate for a snowflake.
	 */
	private double pickRandomX() {
		RandomGenerator rgen = RandomGenerator.getInstance();
		return rgen.nextDouble(0, getWidth() - SNOWFLAKE_SIZE);
	}
	
	/**
	 * Creates and returns a snowflake at a random X coordinate.
	 * 
	 * @return A snowflake at a random X coordinate.
	 */
	private GOval makeSnowflake() {
		GOval result = new GOval(SNOWFLAKE_SIZE, SNOWFLAKE_SIZE);
		result.setFilled(true);
		result.setColor(SNOWFLAKE_COLOR);
		
		/* Reposition the snowflake at the given position. */
		result.setLocation(pickRandomX(), 0);
		return result;
	}
	
	/**
	 * Returns whether the given snowflake has hit the ground.
	 * 
	 * @param snowflake The snowflake to test.
	 * @return Whether the snowflake has hit the ground.
	 */
	private boolean hasHitGround(GOval snowflake) {
		return snowflake.getY() + snowflake.getHeight() >= getHeight();
	}
	
	/**
	 * Returns whether the given snowflake has hit another
	 * object on the screen.
	 * 
	 * @param snowflake The snowflake to test.
	 * @return Whether the snowflake has hit some other object.
	 */
	private boolean hasHitSomething(GOval snowflake) {
		GObject hit = getElementAt(snowflake.getX() + snowflake.getWidth() / 2.0,
				                   snowflake.getY() + snowflake.getHeight());		
		return hit != snowflake;
	}
	
	/**
	 * Simulates the effect of one snowflake falling.
	 */
	private void dropOneSnowflake() {
		GOval snowflake = makeSnowflake();
		add(snowflake);
		
		/* To avoid issues with collision detection, reorder the
		 * snowflake to be at the back.
		 */
		snowflake.sendToBack();
		
		while (!hasHitGround(snowflake) && !hasHitSomething(snowflake)) {
			snowflake.move(0, Y_VELOCITY);
			pause(PAUSE_TIME);
		}
	}
	
	/**
	 * On a mouse click, places a rectangular block at the spot
	 * of the click.
	 */
	public void mouseClicked(MouseEvent e) {
		GRect rect = new GRect(e.getX() - BLOCK_WIDTH / 2.0,
				               e.getY() - BLOCK_HEIGHT / 2.0,
				               BLOCK_WIDTH, BLOCK_HEIGHT);
		rect.setFilled(true);
		add(rect);
	}
}
